import uuid
from flask import Flask, request, jsonify
from functools import wraps
import json
import base64
from collections import defaultdict
import threading

from .blackbox import generate_function
from .util import get_url


class GameServer:
    def __init__(self, token):
        self.name = str(uuid.uuid4())
        self.submission_lock = threading.Lock()
        self.USERNAME = "colabuser"
        self.PASSWORD = "secretcolab"

        self.submissions = defaultdict(lambda: defaultdict(dict))
        try:
            with open("submissions.json", 'r') as f:
                submissions_loaded = json.load(f)
            for team, team_submissions in submissions_loaded.items():
                for time, team_submission in team_submissions.items():
                    self.submissions[team][time] = team_submission
        except FileNotFoundError:
            pass
                
        self.app = Flask(self.name)

        def require_auth(f):
            @wraps(f)
            def decorated(*args, **kwargs):
                auth = request.headers.get("Authorization")
                error = None
                if auth and auth.startswith("Basic "):
                    try:
                        decoded = base64.b64decode(auth.split(" ")[1]).decode("utf-8")
                        user, pw = decoded.split(":")
                        if user == self.USERNAME and pw == self.PASSWORD:
                            return f(*args, **kwargs)
                    except Exception as e:
                        print(e)
                        error = e
                return jsonify({"error": "Unauthorized"}), 401 if error is None else jsonify({'error': str(error)}), 500
            return decorated
        
        @self.app.route("/receive", methods=["POST"])
        @require_auth
        def receive():
            data = request.get_json()
            print("Received data:", data)
            generate_function(data['submission'])
            with self.submission_lock:
                self.submissions[data['team']][data['time']] = data['submission']
                with open("submissions.json", "w") as f:
                    json.dump(dict(self.submissions), f, indent=4)
                print(json.dumps(self.submissions, indent=4))
            return jsonify({"status": "success"})
        self.token = token

    def run(self, force_restart=False):
        server_id = get_url(self.token, force_restart)
        print('Serving', server_id)
        self.app.run(port=5000)



from contextlib import redirect_stderr, redirect_stdout
import cloudpickle as pickle
import base64
from IPython.terminal.interactiveshell import TerminalInteractiveShell
import io


class Blackbox:
    def __init__(self, source, name='play'):
        self.name = name
        self.stdout = io.StringIO()
        self.stderr = io.StringIO()
        self.shell = TerminalInteractiveShell()
        with redirect_stderr(self.stderr), redirect_stdout(self.stdout):
            out = self.shell.run_cell(source)
            if out.error_before_exec is not None:
                raise out.error_before_exec
            if out.error_in_exec is not None:
                raise out.error_in_exec
        if not isinstance(name, str) or name not in self.shell.user_ns:
            raise Exception(f"Your source code must define a function called `{self.name}`, but only has `{self.shell.user_ns.keys()}`")

    def __call__(self, *args, **kwargs):
        input_variable_name = '__magic_input__'
        with redirect_stderr(self.stderr), redirect_stdout(self.stdout):
            self.shell.user_ns[input_variable_name] = args, kwargs
            out = self.shell.run_cell(f'{self.name}(*{input_variable_name}[0], **{input_variable_name}[1])')
            if out.error_before_exec is not None:
                raise out.error_before_exec
            if out.error_in_exec is not None:
                raise out.error_in_exec
            return out.result


class Plackbox:
    def __init__(self, source):
        self.foo = pickle.loads(source)
        self.stdout = io.StringIO()
        self.stderr = io.StringIO()

    def __call__(self, *args, **kwargs):
        with redirect_stderr(self.stderr), redirect_stdout(self.stdout):
            return self.foo(*args, **kwargs)


def generate_function(data):
    if data[0] == 'p':
        return pickle.loads(base64.b64decode(data[1:]))
    else:
        return Blackbox(data[1:])



from gr_tradinggame.coding.client import Client

server = '52bc-34-16-234-11'


client = Client('Team A', server=server)


def foo(reward: float, lockout: int, t: int, T: int, your_score: float, other_scores: list[float]):
    #z = 1 / lockout
    #z = 1 / remaining_rounds
    #z = other_scores[0]
    #return random.random() > 0.5
    threshold = 0.1
    if lockout <= 0:
        threshold = 0
    if t >= T:
        threshold = 0
    #return 1
    return reward > threshold
client.submit(foo)


sub = Client('Team B', threshold=0.1, server=server)

def foo(reward: float, lockout: int, t: int, T: int, your_score: float, other_scores: list[float]):
    #z = 1 / lockout
    #z = 1 / remaining_rounds
    #z = other_scores[0]
    #return random.random() > 0.5
    threshold = 100
    if lockout <= 0:
        threshold = 0
    if t >= T:
        threshold = 0
    #return 1
    return reward > threshold

client.submit(foo)



import base64
import datetime
import random
import traceback
import requests
from requests.auth import HTTPBasicAuth
import cloudpickle as pickle
import timeit

from .blackbox import Plackbox, Blackbox


class Client:
    def __init__(self, team_name, server=None, threshold=0.1):
        self.team_name = team_name
        self.url = "https://" + server + ".ngrok-free.app"
        self.threshold = threshold

    def test(self, source_or_function, verbose=False, hints=False, allow_state=False):
        return self._submit(source_or_function, hints=hints, submit=False, allow_state=allow_state, verbose=verbose)

    def submit(self, source_or_function, allow_state=False):
        return self._submit(source_or_function, hints=False, submit=True, allow_state=allow_state, verbose=False)
        
    def _submit(self, source_or_function, hints, submit, allow_state, verbose):
        if submit and not self.url:
           print('Automatic submission not enabled. Please send an email with your snippet to the organizers.')
           return
        success = True
        test_cases = [
            ((1.5, 5, 1, 10, 1.3, [1.1, 2.3]), "generic inputs", None),
            ((-1.5, 5, 1, 10, 1.3, [1.1, 2.3]), "negative reward", (False, 'It never makes sense to take a negative reward')),
            ((0, 5, 1, 10, 1.3, [1.1, 2.3]), "zero rewards", (False, "It doesn't make sense to accept a lockout in return for no reward")),
            ((1.5, 5, 1, 10, 1.3, []), "playing by yourself", None),
            ((1.5, 5, 1, 10, -1, [-1.1, 2.3]), "negative score", None),
            ((0.0001, 5, 10, 10, 1.3, [1.1, 2.3]), "no time left after this round", (True, 'There is no time left after this round, you should accept any reward you can get')),
            ((0.0001, 0, 1, 10, 1.3, [1.1, 2.3]), "no lockout", (True, 'There is no lockout, you should accept any reward you can get')),
        ]
        if isinstance(source_or_function, str):
            output = source_or_function
            foo = Blackbox(output)
        else:
            output = pickle.dumps(source_or_function)
            foo = Plackbox(output)
        outputs = dict()
        for i, (args, err, hint) in enumerate(test_cases):
            try:
                a = foo(*args)
                if i not in outputs:
                    outputs[i] = a
                if i == 0 and verbose:
                    print(f'Output on input {args} ("{err}"): {a}')
                if a != outputs[i] and not allow_state:
                        success = False
                        print(f'Error on input {args} ("{err}")')
                        print(f'\tYou previously returned `{outputs[i]}` and are now returning `{a}`')
                        print(f"\tIf this is intentional, set `allow_state=True`")
                elif not isinstance(a, bool):
                    success = False
                    print(f'Error on input {args} ("{err}")')
                    print(f'\tYou should return `False` or `True` but you returned `{a}` of type {type(a)}')
                elif hints and hint is not None and a != hint[0]:
                    success = False
                    print(f'Warning on input {args} ("{err}")')
                    print(f'\t{hint[1]}')
                
            except Exception as e:
                success = False
                print(f'Error on input {args} ("{err}")')
                tb = e.__traceback__.tb_next
                stack = ''.join(['\t' + x for x in traceback.format_exception(type(e), e, tb)])
                print(stack)
        if not success:
            return
        tic = timeit.default_timer()
        for j in range(1_000):
            T = int(1_000_000 * random.random())
            t = int(random.random() * T)
            cooldown = int(10 * random.random())
            reward = 100 * random.random()
            score = reward * T * random.random()
            scores = [reward * T * random.random() for _ in range(10)]
            args = (reward, cooldown, t, T,  score, scores)
            try:
                a = foo(*args)
            except Exception as e:
                success = False
                print(f'Error on input {args}')
                tb = e.__traceback__.tb_next
                stack = ''.join(['\t' + x for x in traceback.format_exception(type(e), e, tb)])
                print(stack)
            else:
                if not isinstance(a, bool):
                    success = False
                    print(f'Error on input {args}')
                    print(f'\tYou should return `False` or `True` but returned `{a}` of type {type(a)}')
            if not success:
                return    
        toc = timeit.default_timer()
        if toc - tic > self.threshold:
            print(f'Error: Your code is too slow. Should finish in {self.threshold}s but took {toc - tic:.2f}s')
            return
        if success:
            if submit:
                self.fancy_submit(source_or_function)
            else:
                print("All tests passed. You're ready to call `submit`\n")

    def fancy_submit(self, output):
        if isinstance(output, str):
          submission = 's' + output
        else:
          submission = 'p' + base64.b64encode(pickle.dumps(output)).decode('utf-8')
        j = {"team": self.team_name, 'time': str(datetime.datetime.now()), 'submission': submission}
        response = requests.post(
            self.url + '/receive',
            json=j,
            auth=HTTPBasicAuth("colabuser", "secretcolab"),
        )
        response.raise_for_status()
        print('Submission successful: ', j['team'], j['time'])



import time
import requests
import os
import signal
import subprocess
import requests


def get_url(token, force_restart):
    def kill_and_restart():
        subprocess.run(["pkill", "ngrok"])
        # ngrok.set_auth_token(token)
        # ngrok_connection = ngrok.connect(5000)
        subprocess.Popen(['ngrok', 'http', '5000', '--authtoken', token], start_new_session=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        time.sleep(5)

    def try_get_url():
        response = requests.get("http://localhost:4040/api/tunnels")
        tunnels = response.json()['tunnels']
        public_url = tunnels[0]['public_url']
        return public_url.split('//')[1].split('.ngrok-free.app')[0]

    if force_restart:
        kill_and_restart()
    try:
        return try_get_url()
    except Exception:
        kill_and_restart()
        return try_get_url()
